package api

import (
	"fmt"
	"math/big"
	"strconv"
	"time"
	"validator/claim_monitor"
	"validator/core"

	"github.com/ethereum/go-ethereum/common"
	"github.com/gin-contrib/cors"
	"github.com/gin-gonic/gin"
	"github.com/sirupsen/logrus"
)

var validator *core.Validator
var claimMonitorCli *claim_monitor.Client
var log = logrus.WithField("module", "api")

func getWorkload(c *gin.Context) {
	// 每天的workload获取
	dateList := make([]string, 0)
	if err := c.ShouldBindJSON(&dateList); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}

	timestampList := make([]uint64, 0)
	for _, date := range dateList {
		t, err := time.Parse("2006-01-02", date)
		if err != nil {
			c.JSON(200, withError(DateOutOfRange))
			return
		}
		utcNow := time.Now().UTC()
		if t.After(time.Date(utcNow.Year(), utcNow.Month(), utcNow.Day(), 0, 0, 0, 0, time.UTC).Add(time.Hour * 24)) {
			c.JSON(200, withError(DateOutOfRange))
			return
		}

		timestampList = append(timestampList, uint64(t.Unix()))
	}

	ret, err := validator.GetWorkload(timestampList)
	if err != nil {
		log.WithError(err).Error("failed to get workload")
		c.JSON(200, withError(InternalError))
		return
	}

	c.JSON(200, withSuccess(ret))
}

func getReward(c *gin.Context) {
	// 在这里获取全部，再通过api获取已领取的
	tmp := struct {
		Address string `json:"address"`
	}{}
	if err := c.ShouldBindJSON(&tmp); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}
	if !common.IsHexAddress(tmp.Address) {
		c.JSON(200, withError(InvalidParams))
		return
	}
	reward := validator.GetReward(common.HexToAddress(tmp.Address))
	ether := big.NewInt(1000000000000000000)

	floatReward, _ := new(big.Float).Quo(new(big.Float).SetInt(reward), new(big.Float).SetInt(ether)).Float64()

	claimedReward, err := claimMonitorCli.GetClaimedAmount(tmp.Address)
	if err != nil {
		log.WithError(err).Error("failed to get claimed reward")
		c.JSON(200, withError(InternalError))
		return
	}
	floatClaimedReward, _ := strconv.ParseFloat(claimedReward, 64)

	c.JSON(200, withSuccess(gin.H{
		"total":     fmt.Sprintf("%.6f", floatReward),
		"claimed":   claimedReward,
		"unclaimed": fmt.Sprintf("%.6f", floatReward-floatClaimedReward),
	}))
}

func getRecord(c *gin.Context) {
	// 通过api获取
	tmp := struct {
		Address  string `json:"address"`
		Page     int    `json:"page"`
		PageSize int    `json:"pageSize"`
	}{}

	if err := c.ShouldBindJSON(&tmp); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}
	if !common.IsHexAddress(tmp.Address) {
		c.JSON(200, withError(InvalidParams))
		return
	}
	data, err := claimMonitorCli.GetClaimRecord(tmp.Address, tmp.Page, tmp.PageSize)
	if err != nil {
		log.WithError(err).Error("failed to get claim record")
		c.JSON(200, withError(InternalError))
		return
	}
	c.Data(200, "application/json", data)
}

func getWithdrawProof(c *gin.Context) {
	tmp := struct {
		Address string `json:"address"`
	}{}
	if err := c.ShouldBindJSON(&tmp); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}
	if !common.IsHexAddress(tmp.Address) || len(tmp.Address) != 42 {
		c.JSON(200, withError(InvalidParams))
		return
	}

	amount, proofs := validator.GetMerkleProof(common.HexToAddress(tmp.Address), "")

	c.JSON(200, withSuccess(gin.H{
		"amount": amount,
		"proofs": proofs,
	}))
}

func getMerkleNodes(c *gin.Context) {
	tmp := struct {
		Date     string `json:"date"`
		Depth    uint   `json:"depth"`
		RootHash string `json:"rootHash"`
	}{}
	if err := c.ShouldBindJSON(&tmp); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}

	_, err := time.Parse(time.DateOnly, tmp.Date)
	if err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}

	if tmp.Depth == 0 {
		tmp.Depth = 3
	}

	nodes := validator.GetDailyMerkleNodes(tmp.Date, int(tmp.Depth), common.HexToHash(tmp.RootHash))
	c.JSON(200, withSuccess(nodes))
}

func getMerkleSumNodes(c *gin.Context) {
	tmp := struct {
		Date     string `json:"date"`
		Depth    uint   `json:"depth"`
		RootHash string `json:"rootHash"`
	}{}
	if err := c.ShouldBindJSON(&tmp); err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}

	_, err := time.Parse(time.DateOnly, tmp.Date)
	if err != nil {
		c.JSON(200, withError(InvalidParams))
		return
	}

	if tmp.Depth == 0 {
		tmp.Depth = 3
	}

	nodes, vals := validator.GetDailyMerkleSumNodes(tmp.Date, int(tmp.Depth), common.HexToHash(tmp.RootHash))
	c.JSON(200, withSuccess(gin.H{
		"nodes": nodes,
		"vals":  vals,
	}))
}

func withSuccess(data interface{}) interface{} {
	return map[string]interface{}{
		"code": 0,
		"msg":  "ok",
		"data": data,
	}
}

func withError(msg string) interface{} {
	return map[string]interface{}{
		"code":  1,
		"msg":   "",
		"error": msg,
	}
}

func StartServer(listenAddress string, v *core.Validator, c *claim_monitor.Client) {
	validator = v
	claimMonitorCli = c

	engine := gin.Default()
	engine.Use(cors.Default())
	initRouter(engine)

	log.WithField("listen", listenAddress).Info("start api server")

	err := engine.Run(listenAddress)
	if err != nil {
		log.WithError(err).Fatal("failed to start api server")
	}
}
